import os
import subprocess

ROOT = os.path.realpath(os.path.dirname(__file__) + '/..') + '/'
os.chdir(ROOT+'tests')


def test_hitemp_single_zip(capfd):
    subprocess.call(['repack', 'hitemp_repack_single_zip.cfg'])
    capfd = capfd.readouterr()
    assert """Reading: '02_03750-04000_HITEMP2010.zip'.
  Flagging lines at  500 K:
  Compression rate:       76.13%,     51,032/   213,769 lines.
  Flagging lines at  700 K:
  Compression rate:       67.99%,     68,431/   213,769 lines.
  Total compression rate: 66.28%,     72,084/   213,769 lines.

With a threshold strength factor of 0.01,
kept a total of 72,084 line transitions out of 213,769 lines.

Successfully rewriten hitran line-transition info into:
  'CO2_hitran_2.5-2.6um_500-700K_lbl.dat' and
  'CO2_hitran_2.5-2.6um_500-700K_continuum.dat'.""" in capfd.out


def test_hitemp_single_unzip(capfd):
    # Unzip files before repacking:
    subprocess.call(['unzip', '02_03750-04000_HITEMP2010.zip'])
    subprocess.call(['repack', 'hitemp_repack_single_unzip.cfg'])
    capfd = capfd.readouterr()
    assert """Reading: '02_3750-4000_HITEMP2010.par'.
  Flagging lines at  500 K:
  Compression rate:       76.13%,     51,032/   213,769 lines.
  Flagging lines at  700 K:
  Compression rate:       67.99%,     68,431/   213,769 lines.
  Total compression rate: 66.28%,     72,084/   213,769 lines.

With a threshold strength factor of 0.01,
kept a total of 72,084 line transitions out of 213,769 lines.

Successfully rewriten hitran line-transition info into:
  'CO2_hitran_2.5-2.6um_500-700K_lbl.dat' and
  'CO2_hitran_2.5-2.6um_500-700K_continuum.dat'.""" in capfd.out
    # Teardown:
    os.remove('02_3750-4000_HITEMP2010.par')


def test_hitemp_two_files(capfd):
    subprocess.call(['repack', 'hitemp_repack_two.cfg'])
    capfd = capfd.readouterr()
    assert """Reading: '02_03750-04000_HITEMP2010.zip'.
  Flagging lines at  500 K:
  Compression rate:       76.13%,     51,032/   213,769 lines.
  Flagging lines at  700 K:
  Compression rate:       67.99%,     68,431/   213,769 lines.
  Total compression rate: 66.28%,     72,084/   213,769 lines.

Reading: '02_04000-04500_HITEMP2010.zip'.
  Flagging lines at  500 K:
  Compression rate:       46.14%,     74,466/   138,258 lines.
  Flagging lines at  700 K:
  Compression rate:       18.99%,    111,996/   138,258 lines.
  Total compression rate: 18.73%,    112,367/   138,258 lines.

With a threshold strength factor of 0.01,
kept a total of 184,451 line transitions out of 352,027 lines.

Successfully rewriten hitran line-transition info into:
  'CO2_hitran_2.2-2.6um_500-700K_lbl.dat' and
  'CO2_hitran_2.2-2.6um_500-700K_continuum.dat'.""" in capfd.out


def test_hitemp_single_chunks(capfd):
    subprocess.call(['repack', 'hitemp_repack_single_chunks.cfg'])
    capfd = capfd.readouterr()
    assert """Reading: '02_03750-04000_HITEMP2010.zip'.
  Flagging lines at  500 K (chunk 1/3):
  Compression rate:       83.40%,     11,830/    71,256 lines.
  Flagging lines at  700 K:
  Compression rate:       78.32%,     15,446/    71,256 lines.
  Total compression rate: 76.31%,     16,884/    71,256 lines.

  Flagging lines at  500 K (chunk 2/3):
  Compression rate:       77.32%,     16,163/    71,256 lines.
  Flagging lines at  700 K:
  Compression rate:       68.80%,     22,235/    71,256 lines.
  Total compression rate: 67.00%,     23,517/    71,256 lines.

  Flagging lines at  500 K (chunk 3/3):
  Compression rate:       67.65%,     23,050/    71,257 lines.
  Flagging lines at  700 K:
  Compression rate:       56.81%,     30,775/    71,257 lines.
  Total compression rate: 55.50%,     31,708/    71,257 lines.

With a threshold strength factor of 0.01,
kept a total of 72,109 line transitions out of 213,769 lines.

Successfully rewriten hitran line-transition info into:
  'CO2_hitran_2.5-2.6um_500-700K_lbl.dat' and
  'CO2_hitran_2.5-2.6um_500-700K_continuum.dat'.""" in capfd.out




# Copyright (c) 2017-2020 Patricio Cubillos and contributors.
# repack is open-source software under the MIT license (see LICENSE).

# repack version:
repack_VER =  1  # Major version
repack_MIN =  4  # Minor version
repack_REV =  0  # Revision



#! /usr/bin/env python

# Copyright (c) 2017-2020 Patricio Cubillos and contributors.
# repack is open-source software under the MIT license (see LICENSE).

import sys
import warnings
import time
from datetime import date

import repack as rep


banner = ':'*70

def main():
    warnings.simplefilter("ignore", RuntimeWarning)
    if len(sys.argv) != 2:
        print(f"\n{banner}\n  Error: Wrong usage.\n{banner}\n")
        sys.exit(0)

    print(f"\n{banner}\n"
           "  repack: line-transition data compression.\n"
          f"  Version {rep.__version__}.\n"
          f"  Copyright (c) 2017-{date.today().year} Patricio Cubillos.\n"
           "  repack is open-source software under the MIT license.\n"
          f"{banner}\n\n")

    print(f"Start: {time.ctime()}")
    cfile = sys.argv[1]
    rep.repack(cfile)
    print(f"End: {time.ctime()}")

if __name__ == "__main__":
    main()



# Copyright (c) 2017-2020 Patricio Cubillos and contributors.
# repack is open-source software under the MIT license (see LICENSE).

__all__ = [
    "kB",
    "amu",
    "e",
    "me",
    "N0",
    "nano",
    "C1",
    "C2",
    "C3",
    "ROOT",
    ]

import os

import scipy.constants as sc


# Boltzmann constant in erg K-1:
kB  = sc.k * 1e7
# Unified atomic mass in g:
amu = sc.physical_constants["unified atomic mass unit"][0] * 1e3
# Elementary charge in statcoulomb (cm3/2 g1/2 s-1):
e  = 4.803205e-10
# Electron mass in g:
me = sc.m_e * 1e3
# Amagat in molecules cm-3:
N0 = sc.physical_constants[
        "Loschmidt constant (273.15 K, 101.325 kPa)"][0] * 1e-6

# One nanometer in centimeters:
nano = 1e-7

# Other constructed constants:
C1  = 4.0 * sc.epsilon_0 * sc.m_e * sc.c**2 / sc.e**2 * 0.01  # cm-1
C2  = sc.h * (sc.c * 100.0) / sc.k                            # cm K-1
C3  = sc.pi * e**2 / (me * (100*sc.c)**2)                     # cm

ROOT = os.path.realpath(os.path.dirname(__file__) + "/../..") + "/"



import os
import pytest

import numpy as np

import repack.utils as u

ROOT = os.path.realpath(os.path.dirname(__file__) + '/..') + '/'
os.chdir(ROOT+'tests')


@pytest.mark.parametrize('zip', ['', '.bz2'])
def test_parse_file_exomol(zip):
    tfile = '1H2-16O__POKAZATEL__00400-00500.trans' + zip
    info = u.parse_file(tfile, 'exomol')
    suffix, molecule, isotope, pffile, sfile = info
    assert suffix == '__00400-00500'
    assert molecule == 'H2O'
    assert isotope == '116'
    assert pffile.endswith('1H2-16O__POKAZATEL.pf')
    assert sfile.endswith('1H2-16O__POKAZATEL.states' + zip)


def test_parse_file_exomol_no_suffix():
    info = u.parse_file('1H-12C-14N__Harris.trans', 'exomol')
    suffix, molecule, isotope, pffile, sfile = info
    assert suffix == ''
    assert molecule == 'HCN'
    assert isotope == '124'
    assert pffile.endswith('1H-12C-14N__Harris.pf')
    assert sfile.endswith('1H-12C-14N__Harris.states')


def test_parse_file_hitran():
    info = u.parse_file('02_03750-04000_HITEMP2010.zip', 'hitran')
    suffix, molecule, isotope, pffile, sfile = info
    assert suffix == '03750'
    assert molecule == 'CO2'
    assert isotope is None
    assert pffile is None
    assert sfile is None


def test_parse_file_kurucz():
    info = u.parse_file('tioschwenke.bin', 'kurucz')
    suffix, molecule, isotope, pffile, sfile = info
    assert suffix == ''
    assert molecule == 'TiO'
    assert isotope is None
    assert pffile is None
    assert sfile is None


def test_read_pf_exomol():
    temp, pf = u.read_pf('14N-1H3__BYTe.pf', 'exomol')
    np.testing.assert_equal(temp, np.linspace(1.0, 1600.0, 1600))
    assert np.shape(pf) == (1600,)
    assert pf[ 0] ==     3.8341
    assert pf[-1] == 76420.3468


def test_read_pf_pyrat():
    temp, pf, iso = u.read_pf('PF_tips_CO2.dat', 'pyrat')
    np.testing.assert_equal(temp, np.linspace(70.0, 3000.0, 294))
    assert np.shape(pf) == (11,294)
    assert pf[ 0, 0] == 62.559
    assert pf[10, 293] == 3309200.0
    np.testing.assert_equal(iso, np.array(['626', '636', '628', '627',
        '638', '637', '828', '728', '727', '838', '837']))


@pytest.mark.parametrize('db, molecule, isotope',
   [('1H2-16O__POKAZATEL__00400-00500.trans.bz2',   'H2O', '116'),
    ('1H-2H-16O__VTT__00250-00500.trans.bz2',       'H2O', '126'),
    ('12C-16O2__HITEMP.trans',                      'CO2', '266'),
    ('12C-16O-18O__Zak.par',                        'CO2', '268'),
    ('12C-1H4__YT10to10__01100-01200.trans.bz2',    'CH4', '21111'),
    ('12C-1H3-2H__MockName__01100-01200.trans.bz2', 'CH4', '21112')])
def test_get_exomol_mol(db, molecule, isotope):
    mol, iso = u.get_exomol_mol(db)
    assert mol == molecule
    assert iso == isotope


def test_read_lbl():
    wn, elow, gf, iiso = u.read_lbl('CO2_hitran_2.2-2.6um_500-700K_lbl.dat')
    nlines = 184451
    assert len(wn)   == nlines
    assert len(elow) == nlines
    assert len(gf)   == nlines
    assert len(iiso) == nlines
    assert wn[ 0] == 3750.00024
    assert wn[-1] == 4499.9982
    assert elow[ 0] == 6971.2258
    assert elow[-1] == 6302.3546
    np.testing.assert_approx_equal(gf[ 0], 6.27118677e-06)
    np.testing.assert_approx_equal(gf[-1], 5.64550190e-09)
    np.testing.assert_equal(np.unique(iiso),
        np.array([626, 627, 628, 636, 638, 828]))




# Copyright (c) 2017-2020 Patricio Cubillos and contributors.
# repack is open-source software under the MIT license (see LICENSE).

from .utilities import *
from .cutils    import *

__all__ = (
    utilities.__all__
  + ["flag", "continuum", "flip"]
    )

# Clean up top-level namespace--delete everything that isn't in __all__
# or is a magic attribute, and that isn't a submodule of this package
for varname in dir():
    if not ((varname.startswith('__') and varname.endswith('__')) or
            varname in __all__):
        del locals()[varname]
del(varname)



# Copyright (c) 2017-2020 Patricio Cubillos and contributors.
# repack is open-source software under the MIT license (see LICENSE).

__all__ = [
    'parser',
    'repack',
    ]

import sys
import os
import struct
import subprocess
import configparser
import multiprocessing as mp

import numpy as np
import scipy.interpolate as sip
import scipy.constants as sc

from . import utils     as u
from . import constants as c


def parser(cfile):
    """
    Parse input arguments from configuration file.

    Parameters
    ----------
    cfile: String
        A configuration file.

    Returns
    -------
    files: List of strings
        List with line-transition file names.
    outfile: String
        Output file root name.
    tmin: Float
        Minimum temperature to sample.
    tmax: Float
        Maximum temperature to sample.
    dtemp: Float
        Temperature sampling rate.
    wnmin: Float
        Minimum wavenumber (cm-1) to sample.
    wnmax: Float
        Maximum wavenumber (cm-1) to sample.
    dwn: Float
        Wavenumber sampling rate (cm-1).
    sthresh: Float
        Threshold tolerance level for weak/strong lines.
    pffile: String
        Input partition-function file (for HITRAN input dbtype).
    chunksize: Integer
        Maximum size of chunks to read.
    ncpu: Integer
        Number of parallel CPUs to use.
    """
    config = configparser.ConfigParser()
    config.read([cfile])
    section = "REPACK"
    # Input line-transition files:
    lblfiles = config.get(section, "lblfiles")
    lblfiles = lblfiles.split()
    # Database type:
    db = config.get(section, "dbtype")
    # Output file:
    outfile = config.get(section, "outfile")
    # Partition-function file:
    pffile = None
    if config.has_option(section, "pffile"):
        pffile = config.get(section, "pffile")
    # Max chunk size:
    chunksize = 5000000
    if config.has_option(section, "chunksize"):
        chunksize = config.getint(section, "chunksize")

    ncpu = 1
    if config.has_option(section, "ncpu"):
        ncpu = config.getint(section, "ncpu")
    ncpu = np.clip(ncpu, 1, mp.cpu_count()-1)

    # Temperature sampling:
    tmin  = config.getfloat(section, "tmin")
    tmax  = config.getfloat(section, "tmax")
    dtemp = config.getfloat(section, "dtemp")

    # Wavenumber sampling:
    wnmin = config.getfloat(section, "wnmin")
    wnmax = config.getfloat(section, "wnmax")
    dwn   = config.getfloat(section, "dwn")

    # Line-strength threshold:
    sthresh = config.getfloat(section, "sthresh")

    return lblfiles, db, outfile, tmin, tmax, dtemp, wnmin, wnmax, dwn, \
        sthresh, pffile, chunksize, ncpu


def worker(input, output):
    """
    Doc me!
    """
    for args in iter(input.get, 'STOP'):
        wn, gf, Elow, iiso, tmin, tmax, zmin, zmax, imass, \
            iratio, sthresh, idx = args

        # Low temperature line flagging:
        s = (gf*iratio[iiso]/zmin[iiso] *
             np.exp(-c.C2*Elow/tmin) * (1-np.exp(-c.C2*wn/tmin)))
        alphad = wn/(100*sc.c) * np.sqrt(2.0*c.kB*tmin / (imass[iiso]*c.amu))
        # Line-strength sorted in descending line-trength order:
        isort = np.argsort(alphad/s)
        flag = np.ones(len(iiso), bool)
        u.flag(flag, wn, s/alphad, isort, alphad, sthresh)

        # High temperature line flagging:
        s = (gf*iratio[iiso]/zmax[iiso] *
             np.exp(-c.C2*Elow/tmax) * (1-np.exp(-c.C2*wn/tmax)))
        alphad = wn/(100*sc.c) * np.sqrt(2.0*c.kB*tmax / (imass[iiso]*c.amu))
        isort = np.argsort(alphad/s)
        flag2 = np.ones(len(iiso), bool)
        u.flag(flag2, wn, s/alphad, isort, alphad, sthresh)

        output.put((flag,flag2, wn, gf, Elow, iiso, idx))


def repack(cfile):
    """
    Re-pack line-transition data into lbl data for strong lines and
    continuum data for weak lines.

    Parameters
    ----------
    cfile: String
        A repack configuration file.
    """
    banner = 70 * ":"
    # Parse configuration file:
    args = parser(cfile)
    files, dbtype, outfile, tmin, tmax, dtemp, wnmin, wnmax, dwn, \
        sthresh, pffile, chunksize, ncpu = args

    missing = [file for file in files if not os.path.exists(file)]
    if len(missing) > 0:
        miss_list = '\n  '.join(missing)
        print(f"\n{banner}\n"
               "  File(s) not Found Error: These files are missing:\n"
              f"  {miss_list}"
              f"\n{banner}\n")
        sys.exit(0)

    # Grid sampling for continuum:
    if dwn != 0 and dtemp != 0:
        ntemp = int((tmax-tmin)/dtemp + 1)
        nwave = int((wnmax-wnmin)/dwn + 1)
    else:
        ntemp, nwave = 0, 0

    temperature = np.linspace(tmin, tmax, ntemp)
    wnspec      = np.linspace(wnmin, wnmax, nwave)
    continuum = np.zeros((nwave, ntemp), np.double)

    if dbtype not in ["hitran", "exomol", "kurucz"]:
        print(f"\n{banner}\n  Error: Invalid database ({dbtype}), "
              f"dbtype must be either hitran, exomol, or kurucz.\n{banner}\n")
        sys.exit(0)

    # Parse input files:
    nfiles = len(files)
    suff, mol, isot, pf, states = [], [],  [], [], []
    for dfile in files:
        s, m, iso, p, st = u.parse_file(dfile, dbtype)
        suff.append(s)
        mol.append(m)
        isot.append(iso)
        pf.append(p)
        if st is not None:
            states.append(st)

    # Uncompress states:
    allstates = np.unique(states)
    sdelete, sproc = [], []
    for state in allstates:
        if state.endswith(".bz2"):
            proc = subprocess.Popen(["bzip2", "-dk", state])
            sproc.append(proc)
            sdelete.append(os.path.realpath(state).replace(".bz2", ""))

    if len(np.unique(mol)) > 1:
        print(f"\n{banner}\n"
               "  Error: All input files must correspond to the same molecule."
              f"\n{banner}\n")
        sys.exit(0)
    mol = mol[0]

    z = []  # Interpolator function for partition function per isotope
    if pffile is not None:
        # Read input partition-function file (if given):
        pftemp, partf, isotopes = u.read_pf(pffile, dbtype="pyrat")
        isotopes = list(isotopes)
        for pfvalue in partf:
            z.append(sip.interp1d(pftemp, pfvalue, kind='slinear'))
    else:
        isotopes = list(np.unique(isot))
    niso = len(isotopes)

    # Isotopic abundance ratio and mass:
    iratio, imass = u.read_iso(mol, isotopes, dbtype)

    s = suff[:]  # Make a copy
    # File indices for each wavenumber set:
    wnset = []
    while np.size(s) != 0:
        suffix = s[np.argmin(s)]
        idx = []
        for i in range(nfiles):
            if suff[i] == suffix:
                idx.append(i)
                s.remove(suffix)
        wnset.append(idx)
    nsets = len(wnset)  # Number of wavenumber sets:

    # Number of sets ahead to unzip:
    zbuffer = np.amin([2,nsets])
    tdelete, tproc = [], []
    for b in range(zbuffer):
        tdelete.append([])
        tproc.append([])
        for idx in wnset[b]:
            if files[idx].endswith(".bz2"):
                print(f"Unzipping: '{files[idx]}'.")
                proc = subprocess.Popen(["bzip2", "-dk", files[idx]])
                tproc[b].append(proc)
                tdelete[b].append(files[idx].replace(".bz2", ""))

    for proc in sproc:
        proc.communicate()

    iso = np.zeros(nfiles, int)
    if dbtype == "exomol":
        for i in range(nfiles):
            iso[i] = isotopes.index(isot[i])
        # Read partition-function and states files:
        lblargs = []
        for j in range(niso):
            try:
                i = isot.index(isotopes[j])
            except:
                lblargs.append([None, None, None])  # placeholder
                continue
            # Partition function:
            if pffile is None:
                temp, part = u.read_pf(pf[i], dbtype)
                z.append(sip.interp1d(temp, part, kind='slinear'))
            # States:
            elow, degen = u.read_states(states[i])
            lblargs.append([elow, degen, j])

    else:  #dbtype in ["hitran", "kurucz"]:
        lblargs = [[None, None, None]]  # Trust me

    # Turn isotopes from string to integer data type:
    isotopes = np.asarray(isotopes, int)

    # Set output file names:
    lbl_out  = f"{outfile}_lbl.dat"
    cont_out = f"{outfile}_continuum.dat"
    # Output line-by-line file:
    lblf = open(lbl_out, "wb")

    # Create queues and start worker processes:
    task_queue = mp.Queue()
    done_queue = mp.Queue()
    for i in range(ncpu):
        mp.Process(target=worker, args=(task_queue, done_queue)).start()

    total_lines = 0
    # Read line-by-line files:
    for i in range(nsets):
        # Make sure current files are uncompressed:
        for p in tproc[i]:
            p.communicate()
        # Uncompress following set:
        if zbuffer < nsets:
            tdelete.append([])
            tproc.append([])
            for idx in wnset[zbuffer]:
                if files[idx].endswith(".bz2"):
                    print(f"Unzipping: '{files[idx]}'.")
                    proc = subprocess.Popen(["bzip2", "-dk", files[idx]])
                    tproc[zbuffer].append(proc)
                    tdelete[zbuffer].append(files[idx].replace(".bz2", ""))
            zbuffer += 1

        # Gather data from all files in this wavenumber range:
        lbl, istart, nlines = [], [], []
        for k in range(len(wnset[i])):
            # Initialize lbl object (not reading yet):
            idx = wnset[i][k]
            j = int(iso[wnset[i][k]])
            print(f"Reading: '{files[idx]}'.")
            lbl.append(u.lbl(files[idx], dbtype, *lblargs[j]))
            # Find initial value in range:
            i0 = lbl[k].bs(wnmin, 0,  lbl[k].nlines-1)
            while i0 > 0 and lbl[k].getwn(i0-1) >= wnmin:
                i0 -= 1
            # Find final value in range:
            iN = lbl[k].bs(wnmax, i0, lbl[k].nlines-1)
            while iN < lbl[k].nlines-1 and lbl[k].getwn(iN+1) <= wnmin:
                iN += 1
            istart.append(i0)
            nlines.append(iN-i0+1)

        # Count lines, set target chunk size:
        nchunks = int(np.sum(nlines)/chunksize) + 1
        target  = np.sum(nlines)/nchunks
        chunk   = np.zeros((len(wnset[i]), nchunks+1), int)
        # First and last are easy:
        chunk[:,      0] = istart
        chunk[:,nchunks] = chunk[:,0] + nlines
        # Easy-case split if only one file:
        if len(wnset[i]) == 1:
            chunk[0] = np.linspace(chunk[0,0], chunk[0,-1], nchunks+1)
        else:  # Complicated case:
            wnchunk = np.linspace(wnmin, wnmax, nchunks+1)
            # Intermediate boundaries:
            for n in range(1, nchunks):
                zero = np.sum(chunk[:,n-1])
                wnchunk[n] = u.wnbalance(lbl, wnchunk[n-1], wnmax, target, zero)
                for k in range(len(wnset[i])):
                    chunk[k,n] = lbl[k].bs(
                        wnchunk[n], chunk[k,n-1], chunk[k,nchunks])

        # Proccess chunks:
        for n in range(nchunks):
            gf   = np.array([])
            Elow = np.array([])
            wn   = np.array([])
            iiso = np.array([], int)
            for k in range(len(wnset[i])):
                # Read the LBL files by chunks:
                gfosc, el, wnumber, isoID = lbl[k].read(chunk[k,n:n+2])
                gf   = np.hstack([gf,   gfosc  ])
                Elow = np.hstack([Elow, el     ])
                wn   = np.hstack([wn,   wnumber])
                iiso = np.hstack([iiso, isoID  ])
            # Sort by wavelength:
            asort = np.argsort(wn)
            gf   = gf  [asort]
            Elow = Elow[asort]
            wn   = wn  [asort]
            iiso = iiso[asort]

            zmin = np.array([z[j](tmin) for j in range(niso)])
            zmax = np.array([z[j](tmax) for j in range(niso)])

            args = (wn, gf, Elow, iiso, tmin, tmax, zmin, zmax,
                    imass, iratio, sthresh, n)
            task_queue.put(args)

        collect_wn = []
        collect_gf = []
        collect_Elow = []
        collect_iiso = []
        chunk_idx = []
        for n in range(nchunks):
            flag, flag2, wn, gf, Elow, iiso, idx = done_queue.get()

            # Low temperature line flagging:
            comment = f" (chunk {n+1}/{nchunks})" if nchunks > 1 else ""
            ntotal = len(flag)
            nkept = np.sum(flag)
            print(f"  Flagging lines at {tmin:4.0f} K{comment}:")
            print(f"  Compression rate:       {100.0*(1-nkept/ntotal):.2f}%,"
                  f"  {nkept:9,d}/{ntotal:10,d} lines.")
            # High temperature line flagging:
            nkept = np.sum(flag2)
            print(f"  Flagging lines at {tmax:4.0f} K:")
            print(f"  Compression rate:       {100.0*(1-nkept/ntotal):.2f}%,"
                  f"  {nkept:9,d}/{ntotal:10,d} lines.")
            # Update flag:
            flag |= flag2
            nkept = np.sum(flag)
            print(f"  Total compression rate: {100.0*(1-nkept/ntotal):.2f}%,"
                  f"  {nkept:9,d}/{ntotal:10,d} lines.\n")

            # Store weak lines to continuum file as function of temp:
            for t,T in enumerate(temperature):
                Z = np.array([z[j](T) for j in range(niso)])
                # Line strength in cm molec-1
                s = (c.C3 * gf*iratio[iiso]/Z[iiso] *
                     np.exp(-c.C2*Elow/T) * (1-np.exp(-c.C2*wn/T)))
                # Continuum opacity in cm-1 amagat-1 (instead of cm2 molec-1):
                u.continuum(s[~flag], wn[~flag], continuum[:,t], wnspec)
            # Store strong lines to LBL data file:
            collect_wn.append(wn[flag])
            collect_gf.append(gf[flag])
            collect_Elow.append(Elow[flag])
            collect_iiso.append(iiso[flag])
            chunk_idx.append(idx)

        for n in np.argsort(chunk_idx):
            for wn, gf, Elow, iiso in zip(collect_wn[n], collect_gf[n],
                    collect_Elow[n], collect_iiso[n]):
                lblf.write(struct.pack("dddi", wn, Elow, gf, isotopes[iiso]))

        for k in range(len(wnset[i])):
            lbl[k].close()
        # Delete unzipped sets:
        for f in tdelete[i]:
            os.remove(f)
        total_lines += np.sum(nlines)

    # Close LBL file:
    print(f"With a threshold strength factor of {sthresh},\n"
          f"kept a total of {lblf.tell()/struct.calcsize('dddi'):,.0f} "
          f"line transitions out of {total_lines:,.0f} lines.\n")
    lblf.close()

    for i in range(ncpu):
        task_queue.put('STOP')

    if ntemp != 0:
        # Convert from cm2 molec-1 to cm-1 amagat-1:
        continuum *= c.N0
        # Save Continuum data to file:
        with open(cont_out, "w") as f:
            # Write header:
            f.write(f"@SPECIES\n{mol}\n\n")
            f.write("@TEMPERATURES\n        ")
            for temp in temperature:
                f.write(f" {temp:10.0f}")
            f.write("\n\n")
            # Write the data:
            f.write("# Wavenumber in cm-1, opacity in cm-1 amagat-1:\n")
            f.write("@DATA\n")
            for i in range(nwave):
                f.write(f" {wnspec[i]:12.6f} ")
                for j in range(ntemp):
                    f.write(f" {continuum[i,j]:10.4e}")
                f.write("\n")
    cont_msg = f" and\n  '{cont_out}'" if ntemp != 0 else ""

    print(f"Successfully rewriten {dbtype} line-transition info into:\n"
          f"  '{lbl_out}'{cont_msg}.")

    # Delete unzipped set:
    for f in sdelete:
        os.remove(f)



# Copyright (c) 2017-2020 Patricio Cubillos and contributors.
# repack is open-source software under the MIT license (see LICENSE).

__all__ = [
    "parse_file",
    "read_pf",
    "read_states",
    "lbl",
    "wnbalance",
    "count",
    "read_iso",
    "get_exomol_mol",
    "read_lbl",
    ]

import os
import re
import zipfile
import struct
import itertools

import numpy as np
import scipy.constants as sc

from .. import constants as c


def fopen(filename, mode="r"):
    """
    Find out file compression format (if any) and open the file.

    Parameters
    ----------
    filename: String
        File name.

    Returns
    -------
    file: FILE pointer
    """
    if filename.endswith(".bz2"):
        return open(filename.replace(".bz2", ""), mode)
    elif filename.endswith(".zip"):
        zfile = zipfile.ZipFile(filename, mode)
        # Zip files have to be unzipped to seek them:
        fname = zfile.extract(zfile.namelist()[0])
        zfile.close()
        return open(fname, mode)
    else:
        return open(filename, mode)


def parse_file(lblfile, dbtype):
    """
    Extract info from an Exomol, HITRAN, or Kurucz line-transition filename.

    Parameters
    ----------
    lblfile: String
        An Exomol trans file.
    dbtype: String
        Database type (hitran, exomol, or kurucz).

    Returns
    -------
    suffix: Strings
        Suffix of the trans file.
    molecule: String
        The molecule name.
    isotope: String
        The isotope name.
    pffile: String
        Partition-function file.
    sfile: String
        States file.
    """
    root, file = os.path.split(os.path.realpath(lblfile))
    if dbtype == "exomol":
        # Auxilliary files:
        sfile = file.replace("trans", "states")
        if sfile.count("__") == 2:
            suffix = sfile[sfile.rindex("__"):sfile.index(".")]
            sfile = sfile.replace(suffix, "")
        else:
            suffix = ""
        sfile  = root + "/" + sfile
        pffile = sfile.replace("states", "pf").strip(".bz2")
        # Get info from file name:
        molecule, isotope = get_exomol_mol(file)

    elif dbtype == "hitran":
        hitempID = {"01":"H2O", "02":"CO2", "05":"CO", "08":"NO"}
        molID = file[0:2]
        molecule = hitempID[molID]
        suffix = file[file.find("_")+1:file.rfind(".par")]
        if suffix.find("-") > 0:
            irange = suffix[0:suffix.find("-")]
            suffix = irange.zfill(5)
        # Get them later from the input PF file (required for HITEMP):
        isotope = None
        pffile  = None
        sfile   = None

    elif dbtype == "kurucz":  # TiO is the only available molecule so far
        molecule = "TiO"
        suffix  = ""
        isotope = None
        pffile  = None
        sfile   = None
    return suffix, molecule, isotope, pffile, sfile


def read_pf(pffile, dbtype="exomol"):
    """
    Read an Exomol partition-function file.

    Parameters
    ----------
    pffile: String
        Partition-function file name.
    dbtype: String
        Database type (hitran or exomol).

    Returns
    -------
    temp: 1D float ndarray
        Tabulated list of temperaures.
    pf: 1D/2D float ndarray
        Partition-function values for each temperature.
    isotopes: 1D string ndarray
        The isotope names [returned only for pyrat dbtype].

    Examples
    --------
    >>> import repack.utils as u
    >>> import numpy as np
    >>> # See PF files in: https://github.com/pcubillos/repack/tree/master/tests
    >>> temp, pf = u.read_pf('14N-1H3__BYTe.pf', 'exomol')
    >>> print(temp)
    [1.000e+00 2.000e+00 3.000e+00 ... 1.598e+03 1.599e+03 1.600e+03]
    >>> print(pf)
    [3.83410000e+00 6.78330000e+00 8.21920000e+00 ... 7.60865408e+04
     7.62532828e+04 7.64203468e+04]

    >>> temp, pf, iso = u.read_pf('PF_tips_CO2.dat', 'pyrat')
    >>> with np.printoptions(threshold=10):
    >>>     print(temp)
    [  70.   80.   90. ... 2980. 2990. 3000.]
    >>> print(iso)
    ['626' '636' '628' '627' '638' '637' '828' '728' '727' '838' '837']
    >>> print(pf)
    [[6.2559e+01 7.1483e+01 8.0410e+01 ... 1.0908e+05 1.1054e+05 1.1202e+05]
     [1.2511e+02 1.4296e+02 1.6081e+02 ... 2.3478e+05 2.3794e+05 2.4114e+05]
     [1.3258e+02 1.5150e+02 1.7042e+02 ... 2.3782e+05 2.4101e+05 2.4424e+05]
     ...
     [2.3929e+03 2.7343e+03 3.0758e+03 ... 4.2990e+06 4.3567e+06 4.4150e+06]
     [1.4072e+02 1.6080e+02 1.8089e+02 ... 2.7920e+05 2.8297e+05 2.8679e+05]
     [1.6410e+03 1.8751e+03 2.1094e+03 ... 3.2217e+06 3.2652e+06 3.3092e+06]]
    """
    # Read partition-function file:
    with fopen(pffile) as f:
        lines = f.readlines()

    if dbtype == "exomol":
        # Alocate outputs:
        ntemp = len(lines)
        temp = np.zeros(ntemp, np.double)
        pf   = np.zeros(ntemp, np.double)
        # Extract data:
        for i in np.arange(ntemp):
            temp[i], pf[i] = lines[i].split()
        return temp, pf

    if dbtype == "pyrat":
        # Number of header lines (to skip later when reading the data):
        nskip = 0
        while True:
            line = lines[nskip].strip()
            # Skip blank/empty lines:
            if line == "" or line.startswith('#'):
                pass
            # Read isotopes:
            elif line == "@ISOTOPES":
                isotopes = np.asarray(lines[nskip+1].strip().split())
            # Stop when the tabulated data begins:
            if line == "@DATA":
                nskip += 1
                break
            nskip += 1

        # Number of isotopes:
        niso = len(isotopes)
        # Number of temperature samples:
        ntemp = len(lines) - nskip

        # Allocate output arrays:
        temp = np.zeros(ntemp, np.double)
        pf   = np.zeros((niso, ntemp), np.double)
        # Extract the data:
        for i in np.arange(ntemp):
            info = lines[nskip+i].strip().split()
            temp[i] = info[0]
            pf[:,i] = info[1:]
        return temp, pf, isotopes


def read_states(states):
    """
    Read an Exomol states file.

    Parameters
    ----------
    states: String
       An Exomol states filename.

    Returns
    -------
    elow: 1D float ndarray
       State energy (cm-1).
    g: 1D integer ndarray
       State total statistical degeneracy.
    """
    # Read states file:
    with fopen(states) as f:
        lines = f.readlines()
    nstates = len(lines)
    # Alocate outputs:
    elow = np.zeros(nstates, np.double)  # State energy
    g    = np.zeros(nstates, int)        # State degeneracy (incl. ns)
    # Extract data:
    for i in range(nstates):
        elow[i], g[i] = lines[i].split()[1:3]
    return elow, g


class lbl():
    def __init__(self, lblfile, dbtype, elow, g, iso):
        """
        Parameters
        ----------
        lblfile: String
            A line-transition file.
        dbtype: String
            Database type (hitran or exomol).
        elow: 1D float ndarray
            The states energy (cm-1).
        g: 1D float ndarray
            The states total statistical degeneracy.
        iso: Integer
            The isotope index for this file (for Exomol data).
        """
        self.lblfile = lblfile                    # The file name
        self.dbtype  = dbtype                     # Database type

        if dbtype == "kurucz":
            self.file = fopen(lblfile, "rb")
            self.llen = 16
            self.ratiolog = np.log(1.0 + 1.0/2000000)
            self.tablog = 10.0**(0.001*(np.arange(32769) - 16384))
        else:
            self.file  = fopen(lblfile, "r")  # The actual file
            dummy = self.file.readline()
            self.llen = self.file.tell()      # Length of lines in file
        self.file.seek(0,2)
        self.nlines = self.file.tell() // self.llen  # Number of lines
        self.elow   = elow
        self.g      = g
        self.iso    = iso  # Isotope index

    def bs(self, val, lo, hi):
        """
        Wavenumber binary search on database.

        Parameters
        ----------
        val: Float
            Target wavenumber value (in cm-1).
        lo: Integer
            Initial wavenumber index where to search.
        hi: Integer
            Final wavenumber index where to search.

        Returns
        -------
        index: Integer
            The index of the closest wavenumber entry to val.
        """
        # Out of bounds:
        if val <= self.getwn(0):
            return 0
        if val >= self.getwn(self.nlines-1):
            return self.nlines-1

        # End case:
        if hi-lo <= 1:
            if np.abs(val-self.getwn(hi)) < np.abs(val-self.getwn(lo)):
                return hi
            return lo

        # Half it:
        mid = int(0.5*(hi+lo))
        if self.getwn(mid) > val:
            return self.bs(val, lo, mid)
        return self.bs(val, mid, hi)


    def getwn(self, index):
        """
        Extract wavenumber from file at the requested index.

        Parameters
        ----------
        index: Integer
            The wavenumber index in database to extract.
        Returns
        -------
        wn: Float
            The wavenumber (cm-1) at position index.
        """
        if self.dbtype == "exomol":
            self.file.seek(index*self.llen)
            iup = int(self.file.read(12)) - 1
            self.file.read(1)
            ilo = int(self.file.read(12)) - 1
            return self.elow[iup] - self.elow[ilo]

        elif self.dbtype == "hitran":
            self.file.seek(index*self.llen + 3)
            return float(self.file.read(12))

        elif self.dbtype == "kurucz":
            # Note I'm reversing the indexing because this DB has decreasing wn:
            self.file.seek((self.nlines-index-1)*self.llen)
            # 4 = struct.calcsize("i")
            iw = struct.unpack('i', self.file.read(4))[0]
            return 1.0/(np.exp(iw * self.ratiolog) * c.nano)


    def read(self, chunk):
        """
        Read a chunk of line transitions.

        Parameters
        ----------
        chunk: Two-element tuple
            Initial and final indices of the chunk to read.

        Returns
        -------
        gf: 1D float ndarray
            Transition weighted oscillator strength (unitless).
        Elow: 1D float ndarray
            Transition lower-state energy (cm-1).
        wn: 1D float ndarray
            Transition wavenumber (cm-1).
        isoID: 1D integer ndarray
            isotope ID (for hitran dbtype).
        """
        # Go to beginning of chunk:
        self.file.seek(chunk[0]*self.llen)
        nlines = chunk[1] - chunk[0]
        # Extract info:
        if self.dbtype == "exomol":
            iup = np.zeros(nlines, int)
            ilo = np.zeros(nlines, int)
            A21 = np.zeros(nlines, np.double)
            for i in range(nlines):
                  line = self.file.readline()
                  iup[i] = line[ 0:12]
                  ilo[i] = line[13:25]
                  A21[i] = line[26:36]
            iup -= 1
            ilo -= 1
            # Compute values:
            wn   = self.elow[iup] - self.elow[ilo]
            gf   = self.g[ilo] * A21 * c.C1 / (8.0*np.pi*100*sc.c) / wn**2.0
            Elow = self.elow[ilo]
            isoID = np.tile(self.iso, np.size(wn))

        elif self.dbtype == "hitran":
            isoID = np.zeros(nlines,       int)
            wn    = np.zeros(nlines, np.double)
            A21   = np.zeros(nlines, np.double)
            Elow  = np.zeros(nlines, np.double)
            g     = np.zeros(nlines, np.double)
            for i in range(nlines):
                line = self.file.readline()
                isoID[i] = line[  2: 3]
                wn   [i] = line[  3:15]
                A21  [i] = line[ 25:35]
                Elow [i] = line[ 45:55]
                g    [i] = line[155:self.llen]
            gf = g * A21 * c.C1 / (8.0*np.pi*100*sc.c) / wn**2.0
            isoID = (isoID - 1) % 10

        elif self.dbtype == "kurucz":
            iw   = np.zeros(nlines, int)
            ieli = np.zeros(nlines, np.short)
            ielo = np.zeros(nlines, np.short)
            igf  = np.zeros(nlines, np.short)
            i = 0
            while i < nlines:
                idx = self.nlines - 1 - (chunk[0]+i)  # Reverse indexing again
                self.file.seek(idx*self.llen)
                # 10 = struct.calcsize("ihhh")
                iw[i], ieli[i], ielo[i], igf[i] = struct.unpack(
                    'ihhh', self.file.read(10))
                i += 1
            wn    = 1.0/(np.exp(iw * self.ratiolog) * c.nano)
            gf    = self.tablog[igf]
            isoID = np.abs(ieli) - 8950
            Elow  = self.tablog[ielo]

        return gf, Elow, wn, isoID


    def close(self):
        if self.lblfile.endswith(".zip"):
            # Remove unzipped files:
            with zipfile.ZipFile(self.lblfile, "r") as zfile:
                os.remove(zfile.namelist()[0])
        self.file.close()


def wnbalance(lbls, wnmin, wnmax, targetsize, zero=0, tol=0.01):
    """
    Binary search of wavenumber value (wntarget) such that there are
    targetsize+zero line-transitions with wavenumber < wntarget.

    Parameters
    ----------
    lbls: List of lbl objects
        Line by line objects.
    wnmin: Float
        Minimum wavenumber to consider
    wnmax: Float
        Maximum wavenumber to consider
    targetsze: Integer
        The desired number of transitions with wavenumber < wntarget.
    zero: Integer
        Zero offset number of transitions to discount.
    tol: Float
        Fractional tolerance level (w.r.t. targetsize) of to accept wntarget.

    Returns
    -------
    wntarget: Float
        The wavenumber that has targetsize transitions to the left.
    """
    # Middle point between wn boundaries:
    wntarget = 0.5*(wnmax + wnmin)
    # Number of transitions with wn < wntarget:
    ntarget  = count(lbls, wntarget)

    if np.abs(ntarget-zero - targetsize) < tol*targetsize:
        return wntarget

    elif ntarget-zero < targetsize:
        return wnbalance(lbls, wntarget, wnmax,    targetsize, zero, tol)
    else:
        return wnbalance(lbls, wnmin,    wntarget, targetsize, zero, tol)


def count(lbls, wntarget):
    """
    Count the number of line transitions with wavenumber smaller than wntarget.

    Parameters
    ----------
    lbls: List of lbl objects
    wntarget: Float
        The target wavenumber.

    Returns
    -------
    nwave: Integer
        The number of transitions with wavenumber smaller than wntarget.
    """
    nwave = 0
    for lbl in lbls:
        nwave += lbl.bs(wntarget, 0, lbl.nlines-1)
    return nwave


def read_iso(mol, iso, dbtype="exomol", isofile=c.ROOT+"repack/isotopes.dat"):
    """
    Read an isotopes info file.

    Parameters
    ----------
    mol: String
        Molecule name.
    iso: List of strings
        Molecule's isotope name.
    dbtype: String
        Database format for isotope names (hitran or exomol).
    isofile: String
        File containing the isotopic information.

    Returns
    -------
    iratio: List of floats
        Isotopic abudance fraction (unitless).
    imass: List of floats
        Isotopic mass (amu).
    """
    # Alocate outputs:
    iratio = np.zeros(len(iso))
    imass  = np.zeros(len(iso))

    if dbtype in ["exomol", "kurucz"]:  # Kurucz's TiO uses Exomol iso notation
        iiso = 2
    elif dbtype == "hitran":
        iiso = 1
    # Read info file:
    with open(isofile, "r") as f:
        lines = f.readlines()
    # Get values for our molecule/isotopes:
    for i in np.arange(len(lines)):
        if lines[i].startswith("#") or lines[i].strip() == "":
            continue
        info = lines[i].split()
        if info[0] == mol:
            if info[iiso] in iso:
                iratio[iso.index(info[iiso])] = info[3]
                imass [iso.index(info[iiso])] = info[4]
    return iratio, imass


def get_exomol_mol(dbfile):
    """
    Parse an exomol file to extract the molecule and isotope name.

    Parameters
    ----------
    dbfile: String
        An exomol line-list file (must follow ExoMol naming convention).

    Returns
    -------
    molecule: String
        Name of the molecule.
    isotope: String
        Name of the isotope (See Tennyson et al. 2016, jmosp, 327).

    Examples
    --------
    >>> filenames = [
    >>>     '1H2-16O__POKAZATEL__00400-00500.trans.bz2',
    >>>     '1H-2H-16O__VTT__00250-00500.trans.bz2',
    >>>     '12C-16O2__HITEMP.pf',
    >>>     '12C-16O-18O__Zak.par',
    >>>     '12C-1H4__YT10to10__01100-01200.trans.bz2',
    >>>     '12C-1H3-2H__MockName__01100-01200.trans.bz2'
    >>>    ]
    >>> for db in filenames:
    >>>     print(get_exomol_mol(db))
    ('H2O', '116')
    ('H2O', '126')
    ('CO2', '266')
    ('CO2', '268')
    ('CH4', '21111')
    ('CH4', '21112')
    """
    atoms = os.path.split(dbfile)[1].split("_")[0].split("-")
    elements = []
    isotope  = ""
    for atom in atoms:
        match = re.match(r"([0-9]+)([a-z]+)([0-9]*)", atom, re.I)
        N = 1 if match.group(3) == "" else int(match.group(3))
        elements += N * [match.group(2)]
        isotope  += match.group(1)[-1:] * N

    composition = [list(g[1]) for g in itertools.groupby(elements)]
    molecule = "".join([c[0] + str(len(c))*(len(c)>1)
                        for c in composition])

    return molecule, isotope


def read_lbl(lbl_file):
    """
    Read a repack output LBL binary file.

    Parameters
    ----------
    lbl_file: String
        Path to file to read.

    Returns
    -------
    wn: 1D float ndarray
        Line transitions' wavenumber (in cm-1 units).
    elow: 1D float ndarray
        Line transitions' lower-state energy (in cm-1 units).
    gf: 1D float ndarray
        Line transitions' gf value (unitless).
    iiso: 1D integer ndarray
        Line transitions' isotope ID.

    Examples
    --------
    >>> import repack.utils as u
    >>> wn, elow, gf, iiso = u.read_lbl('HCN_exomol_0.3-33um_500-3000K_lbl.dat')
    """
    fmt = 'dddi'
    fmtsize = struct.calcsize(fmt)

    # Number of line transitions:
    f = open(lbl_file, 'rb')
    f.seek(0, 2)
    nlines = f.tell() // fmtsize

    # Allocate output arrays:
    wn   = np.zeros(nlines, np.double)
    elow = np.zeros(nlines, np.double)
    gf   = np.zeros(nlines, np.double)
    iiso = np.zeros(nlines, int)

    # Read file:
    f.seek(0)
    for i in range(nlines):
        wn[i], elow[i], gf[i], iiso[i] = struct.unpack(fmt, f.read(fmtsize))
    f.close()

    return wn, elow, gf, iiso




import os
import subprocess
import pytest

ROOT = os.path.realpath(os.path.dirname(__file__) + '/..') + '/'
os.chdir(ROOT+'tests')


@pytest.mark.parametrize('missfile',
    ['exomol_repack_missing1.cfg',
     'exomol_repack_missing2.cfg'])
def test_missing(capfd, missfile):
    subprocess.call(['repack', missfile])
    capfd = capfd.readouterr()
    assert """File(s) not Found Error: These files are missing:
  14N-1H3__MiSSinG__00200-00300.trans.bz2""" in capfd.out


def test_exomol_single(capfd):
    subprocess.call(['repack', 'exomol_repack_single.cfg'])
    capfd = capfd.readouterr()
    assert """Unzipping: '14N-1H3__BYTe__00100-00200.trans.bz2'.
Reading: '14N-1H3__BYTe__00100-00200.trans.bz2'.
  Flagging lines at  500 K:
  Compression rate:       73.90%,    270,761/ 1,037,545 lines.
  Flagging lines at  700 K:
  Compression rate:       72.20%,    288,406/ 1,037,545 lines.
  Total compression rate: 70.61%,    304,908/ 1,037,545 lines.

With a threshold strength factor of 0.01,
kept a total of 304,908 line transitions out of 1,037,545 lines.

Successfully rewriten exomol line-transition info into:
  'NH3_exomol_050-100um_500-700K_lbl.dat' and
  'NH3_exomol_050-100um_500-700K_continuum.dat'.""" in capfd.out


def test_exomol_two_files(capfd):
    subprocess.call(['repack', 'exomol_repack_two_files.cfg'])
    capfd = capfd.readouterr()
    assert """Unzipping: '14N-1H3__BYTe__00100-00200.trans.bz2'.
Unzipping: '14N-1H3__BYTe__00200-00300.trans.bz2'.
Reading: '14N-1H3__BYTe__00100-00200.trans.bz2'.
  Flagging lines at  500 K:
  Compression rate:       73.90%,    270,761/ 1,037,545 lines.
  Flagging lines at  700 K:
  Compression rate:       72.20%,    288,406/ 1,037,545 lines.
  Total compression rate: 70.61%,    304,908/ 1,037,545 lines.

Reading: '14N-1H3__BYTe__00200-00300.trans.bz2'.
  Flagging lines at  500 K:
  Compression rate:       83.14%,    179,161/ 1,062,896 lines.
  Flagging lines at  700 K:
  Compression rate:       81.82%,    193,215/ 1,062,896 lines.
  Total compression rate: 80.62%,    206,002/ 1,062,896 lines.

With a threshold strength factor of 0.01,
kept a total of 510,910 line transitions out of 2,100,441 lines.

Successfully rewriten exomol line-transition info into:
  'NH3_exomol_033-100um_500-700K_lbl.dat' and
  'NH3_exomol_033-100um_500-700K_continuum.dat'.""" in capfd.out


def test_exomol_two_isotopes(capfd):
    subprocess.call(['repack', 'exomol_repack_two_isotopes.cfg'])
    capfd = capfd.readouterr()
    assert """Unzipping: '14N-1H3__BYTe__00100-00200.trans.bz2'.
Unzipping: '15N-1H3__BYTe-15__00100-00200.trans.bz2'.
Reading: '14N-1H3__BYTe__00100-00200.trans.bz2'.
Reading: '15N-1H3__BYTe-15__00100-00200.trans.bz2'.
  Flagging lines at  500 K:
  Compression rate:       76.63%,    283,467/ 1,212,878 lines.
  Flagging lines at  700 K:
  Compression rate:       75.06%,    302,443/ 1,212,878 lines.
  Total compression rate: 73.26%,    324,263/ 1,212,878 lines.

With a threshold strength factor of 0.01,
kept a total of 324,263 line transitions out of 1,212,878 lines.

Successfully rewriten exomol line-transition info into:
  'NH3_exomol_050-100um_500-700K_lbl.dat' and
  'NH3_exomol_050-100um_500-700K_continuum.dat'.""" in capfd.out


def test_exomol_two_files_two_iso(capfd):
    subprocess.call(['repack', 'exomol_repack_two_two.cfg'])
    capfd = capfd.readouterr()
    assert """Unzipping: '14N-1H3__BYTe__00100-00200.trans.bz2'.
Unzipping: '15N-1H3__BYTe-15__00100-00200.trans.bz2'.
Unzipping: '14N-1H3__BYTe__00200-00300.trans.bz2'.
Unzipping: '15N-1H3__BYTe-15__00200-00300.trans.bz2'.
Reading: '14N-1H3__BYTe__00100-00200.trans.bz2'.
Reading: '15N-1H3__BYTe-15__00100-00200.trans.bz2'.
  Flagging lines at  500 K:
  Compression rate:       76.63%,    283,467/ 1,212,878 lines.
  Flagging lines at  700 K:
  Compression rate:       75.06%,    302,443/ 1,212,878 lines.
  Total compression rate: 73.26%,    324,263/ 1,212,878 lines.

Reading: '14N-1H3__BYTe__00200-00300.trans.bz2'.
Reading: '15N-1H3__BYTe-15__00200-00300.trans.bz2'.
  Flagging lines at  500 K:
  Compression rate:       85.02%,    184,887/ 1,234,214 lines.
  Flagging lines at  700 K:
  Compression rate:       83.77%,    200,355/ 1,234,214 lines.
  Total compression rate: 82.50%,    216,004/ 1,234,214 lines.

With a threshold strength factor of 0.01,
kept a total of 540,267 line transitions out of 2,447,092 lines.

Successfully rewriten exomol line-transition info into:
  'NH3_exomol_033-100um_500-700K_lbl.dat' and
  'NH3_exomol_033-100um_500-700K_continuum.dat'.""" in capfd.out


def test_exomol_single_unzip(capfd):
    # Unzip files before repacking:
    subprocess.call(['bzip2','-dk','14N-1H3__BYTe__00100-00200.trans.bz2'])
    subprocess.call(['bzip2','-dk','14N-1H3__BYTe.states.bz2'])
    subprocess.call(['repack', 'exomol_repack_single_unzip.cfg'])
    capfd = capfd.readouterr()
    assert """Reading: '14N-1H3__BYTe__00100-00200.trans'.
  Flagging lines at  500 K:
  Compression rate:       73.90%,    270,761/ 1,037,545 lines.
  Flagging lines at  700 K:
  Compression rate:       72.20%,    288,406/ 1,037,545 lines.
  Total compression rate: 70.61%,    304,908/ 1,037,545 lines.

With a threshold strength factor of 0.01,
kept a total of 304,908 line transitions out of 1,037,545 lines.

Successfully rewriten exomol line-transition info into:
  'NH3_exomol_050-100um_500-700K_lbl.dat' and
  'NH3_exomol_050-100um_500-700K_continuum.dat'.""" in capfd.out
    os.remove('14N-1H3__BYTe__00100-00200.trans')
    os.remove('14N-1H3__BYTe.states')


def test_exomol_single_chunks(capfd):
    subprocess.call(['repack', 'exomol_repack_single_chunks.cfg'])
    capfd = capfd.readouterr()
    assert """Unzipping: '14N-1H3__BYTe__00100-00200.trans.bz2'.
Reading: '14N-1H3__BYTe__00100-00200.trans.bz2'.
  Flagging lines at  500 K (chunk 1/2):
  Compression rate:       70.42%,    153,438/   518,772 lines.
  Flagging lines at  700 K:
  Compression rate:       68.69%,    162,411/   518,772 lines.
  Total compression rate: 66.97%,    171,335/   518,772 lines.

  Flagging lines at  500 K (chunk 2/2):
  Compression rate:       77.38%,    117,324/   518,773 lines.
  Flagging lines at  700 K:
  Compression rate:       75.71%,    125,996/   518,773 lines.
  Total compression rate: 74.25%,    133,574/   518,773 lines.

With a threshold strength factor of 0.01,
kept a total of 304,909 line transitions out of 1,037,545 lines.

Successfully rewriten exomol line-transition info into:
  'NH3_exomol_050-100um_500-700K_lbl.dat' and
  'NH3_exomol_050-100um_500-700K_continuum.dat'.""" in capfd.out




# Copyright (c) 2017-2020 Patricio Cubillos and contributors.
# repack is open-source software under the MIT license (see LICENSE).

import os
import re
import sys
import setuptools
from setuptools import setup, Extension
from numpy import get_include

topdir = os.path.dirname(os.path.realpath(__file__))
sys.path.append(topdir + "/repack")
import VERSION as v

version = f'{v.repack_VER}.{v.repack_MIN}.{v.repack_REV}'

srcdir = './repack/utils/'  # C-code source folder

# Get all file from source dir:
files = os.listdir(srcdir)
# Filter the results for just the c files:
files = list(filter(lambda x:     re.search('.+[.]c$',     x), files))
files = list(filter(lambda x: not re.search('[.#].+[.]c$', x), files))

inc = [get_include()]
eca = ['-ffast-math']
ela = []

extensions = []
for cfile in files:
    e = Extension('repack.utils.'+cfile.rstrip('.c'),
                  sources=[f"{srcdir}{cfile}"],
                  include_dirs=inc,
                  extra_compile_args=eca,
                  extra_link_args=ela)
    extensions.append(e)

with open('README.md', 'r') as f:
    readme = f.read()

setup(name         = "lbl-repack",
      version      = version,
      author       = "Patricio Cubillos",
      author_email = "patricio.cubillos@oeaw.ac.at",
      url          = "https://github.com/pcubillos/repack",
      packages     = setuptools.find_packages(),
      install_requires = [
                     'numpy>=1.13.3',
                     'scipy>=0.17.1',
                     ],
      include_package_data=True,
      license      = "MIT",
      description  = 'A line-transition data compression package.',
      long_description=readme,
      long_description_content_type="text/markdown",
      include_dirs = inc,
      entry_points={"console_scripts": ['repack = repack.__main__:main']},
      ext_modules  = extensions)




# Copyright (c) 2017-2020 Patricio Cubillos and contributors.
# repack is open-source software under the MIT license (see LICENSE).

from . import utils
from . import constants
from . import VERSION as ver
from .pack import *

__all__ = (
    pack.__all__
  + ['utils']
  + ['constants']
    )

__version__ = "{:d}.{:d}.{:d}".format(ver.repack_VER,
                                      ver.repack_MIN, ver.repack_REV)

# Clean up top-level namespace--delete everything that isn't in __all__
# or is a magic attribute, and that isn't a submodule of this package
for varname in dir():
    if not ((varname.startswith('__') and varname.endswith('__')) or
            varname in __all__):
        del locals()[varname]
del(varname)
